# -*- coding: utf-8 -*-
"""IO methods to read and write config."""
import os

import confuse


def write_to_file(config: confuse.Configuration) -> None:
    """Write updated config to file."""
    config_filename = os.path.join(config.config_dir(), confuse.CONFIG_FILENAME)
    with open(config_filename, "w") as f:
        f.write(config.dump())



# -*- coding: utf-8 -*-
"""Config classes for nessie client."""
from io import StringIO

import confuse
import yaml

from .config_command import process
from .config_parser import build_config
from .io import write_to_file

__all__ = ["build_config", "to_dict", "write_to_file", "process"]


def to_dict(config: confuse.Configuration) -> dict:
    """Convert confuse Configuration object to dict."""
    ys = config.dump(redact=True)
    fd = StringIO(ys)
    dct = yaml.safe_load(fd)
    return dct



# -*- coding: utf-8 -*-
"""Config based auth mechanism."""
import json
import os
import time

import confuse
from confuse import Configuration
from confuse import ConfigValueError
from confuse import NotFoundError
from requests import models
from requests.auth import AuthBase

from .basic import login as base_login


def login(base_url: str, config_dict: confuse.Configuration) -> AuthBase:
    """Return a TokenAuth auth object for standard jwt login."""
    return TokenAuth(base_url, config_dict)


def _login(base_url: str, config_dict: confuse.Configuration) -> str:
    """Log into Nessie using basic auth and looking for config.

    :param base_url: Full Nessie url
    :param config_dict: config dict
    :return: auth token
    """
    username = config_dict["auth"]["username"].get()
    if not username:
        raise RuntimeError("No username available, can't login")
    password = config_dict["auth"]["password"].get()
    if not password:
        raise RuntimeError("No password available, can't login")
    try:
        timeout = config_dict["auth"]["timeout"].get(int)
    except NotFoundError:
        timeout = 10
    try:
        verify = config_dict["verify"].get(bool)
    except NotFoundError:
        verify = 10
    return base_login(base_url, username, password, timeout, verify)


def _write_token(token: str, config_dict: Configuration) -> None:
    for source in config_dict.sources:
        if source.filename and not source.default:
            directory = os.path.dirname(source.filename)
            with open(os.path.join(directory, "auth.json"), "w") as f:
                json.dump(
                    {
                        "token": token,
                        "timestamp": time.time(),
                        "hostname": config_dict["hostname"].get(),
                        "user": config_dict["auth"]["username"].get(),
                    },
                    f,
                )
            return


def _existing_token(config_dict: Configuration) -> str:
    for source in config_dict.sources:
        if source.filename and not source.default:
            directory = os.path.dirname(source.filename)
            if os.path.exists(os.path.join(directory, "auth.json")):
                with open(os.path.join(directory, "auth.json")) as f:
                    authfile = json.load(f)
                    if _is_valid(authfile, config_dict):
                        return authfile["token"]
    raise KeyError


def _is_valid(authfile: dict, config_dict: Configuration) -> bool:
    hostname = authfile["hostname"]
    username = authfile["username"]
    try:
        expected_hostname = config_dict["hostname"].get()
        expected_username = config_dict["auth"]["username"].get()
    except (ConfigValueError, NotFoundError):
        return False
    if hostname != expected_hostname:
        return False
    if username != expected_username:
        return False
    now = time.time()
    then = authfile["timestamp"]
    return (then + 60 * 60 * 10) > now


class TokenAuth(AuthBase):
    """AuthBase override for simple token based auth."""

    def __init__(self: "TokenAuth", base_url: str, config_dict: Configuration) -> None:
        """Fetch token from service."""
        self.config_dict = config_dict
        self.base_url = base_url
        self._token = self._get_token()

    def __call__(self: "TokenAuth", r: models.PreparedRequest) -> models.PreparedRequest:
        """Append Auth Header to request."""
        r.headers["Authorization"] = "Bearer {}".format(self._token)
        return r

    def _get_token(self: "TokenAuth") -> str:
        try:
            return _existing_token(self.config_dict)
        except KeyError:
            token = _login(self.base_url, self.config_dict)
            _write_token(token, self.config_dict)
            return token



# -*- coding: utf-8 -*-
"""Nessie Exceptions."""
import functools
import sys
import traceback
from typing import Any
from typing import Callable

import click
import requests
import simplejson as json

_REMEDIES = {
    "Cannot create an unassigned tag reference": "set a valid reference on which to create this tag. eg `nessie tag tag_name main`"
}


class NessieException(Exception):
    """Base Nessie exception."""

    def __init__(self: "NessieException", msg: str, original_exception: Exception, response: requests.models.Response) -> None:
        """Construct base Nessie Exception."""
        super(NessieException, self).__init__(msg + (": %s" % original_exception))
        self.original_exception = original_exception
        try:
            parsed_response = response.json()

        except:  # NOQA
            parsed_response = dict()

        self.status_code = response.status_code
        self.server_message = parsed_response.get("message", response.text)
        self.server_status = parsed_response.get("status")
        self.server_stack_trace = parsed_response.get("serverStackTrace")
        self.url = response.url

    def json(self: "NessieException") -> str:
        """Dump this exception as a json object."""
        return json.dumps(
            dict(
                server_message=self.server_message,
                server_status=self.server_status,
                server_stack_trace=self.server_stack_trace,
                status_code=self.status_code,
                url=self.url,
                original_exception=traceback.format_tb(self.original_exception.__traceback__),
                msg=" ".join(self.args),
            )
        )


class NessieUnauthorizedException(NessieException):
    """Nessie exception for login error 401."""

    pass


class NessiePermissionException(NessieException):
    """Nessie exception for permission error 403."""

    pass


class NessieNotFoundException(NessieException):
    """Nessie exception for not found error 404."""

    pass


class NessiePreconidtionFailedException(NessieException):
    """Nessie exception for pre-condition failed error 412."""

    pass


class NessieConflictException(NessieException):
    """Nessie exception for conflict error 408."""

    pass


class NessieServerException(NessieException):
    """Nessie exception for server errors 5xx."""

    pass


def error_handler(f: Callable) -> Callable:
    """Wrap a click method to catch and pretty print errors."""

    @functools.wraps(f)
    def wrapper(*args: Any, **kwargs: Any) -> None:
        """Wrapper object."""
        try:
            f(*args, **kwargs)
        except NessieException as e:
            if args[0].json:
                click.echo(e.json())
            click.echo(_format_error(e))
            sys.exit(1)

    return wrapper


def _format_error(e: NessieException) -> str:
    fmt = "{} is {} with status code: {}.\n".format(click.style("Nessie Exception", fg="red"), e, e.status_code)
    if e.server_message in _REMEDIES:
        fmt += "{} {}\n".format(click.style("FIX:", fg="green"), _REMEDIES[e.server_message])
    fmt += "{} {}\n".format(click.style("Original URL:", fg="yellow"), e.url)
    fmt += "{} {}\n".format(click.style("Server status:", fg="yellow"), e.server_status)
    fmt += "{} {}\n".format(click.style("Server message:", fg="yellow"), e.server_message)
    fmt += "{} {}\n".format(click.style("Server traceback:", fg="yellow"), e.server_stack_trace)
    return fmt
